/*
  CPP_NUM_CORES=32
  CPP_PROCESSES_PER_NODE=1
  CPP_PROBLEM=A
*/

/*
** Practica: Concurso de programaciÃ³n paralela - Problema A
** 2011/12/05
** Jonas Koenemann
** 
** This file was aumented by message passing interface and OpenMP.
** I introduced a new function mh_partial, which is executed
** seperately by each node.
** It also contains the pragma directive to distribute the work on
** different threads with shared memory.
**
** The master process divides the work and broadcasts parts
** of the first matrix to the other processes.
** The first matrix is split row-whise in blocks.
** All processes have an equal copy of the second matrix.
**
** Both steps, formarestructura and multiplication, are parallelized,
** whereas only the mulptiplication part is distributed between different
** threads. Number of threads is set to 8.
**
** Finally all processes send their part of the solution to the 
** master process.
*/

/* 
University of La Laguna
time: 950 ms
speedup: 19.5
*/

#include <stdlib.h>
#include <mpi.h>
#include <omp.h>

/**
*
* remember on which positions have values different from 0
* stop marker is value of t
* 
* int t: size
* int *estructura: structure array to write in
* double *m: matrix to read from
* int ld: spacing to distinguish between row and columns reading (either 1 or t)
*
*/
void formarestructura(int t,int *estructura,double *m,int ld)
{
  int i,pos=0,ceros=1;
  
  for(i=0;i<t;i++)
  {
    if(ceros==1 && m[i*ld]!=0.)
    {
      estructura[pos]=i;
      pos++;
      ceros=0;
    }
    else if(ceros==0 && m[i*ld]==0.)
    {
      estructura[pos]=i-1;
      pos++;
      ceros=1;
    }
  }   
  if(ceros==0)
  {
    estructura[pos]=t-1;
    pos++;
  }
  estructura[pos]=t;
}




void mh_partial(int *filasa, int *columnasb, double *a, double *b, double *c, int t,
                  int num_rows) 
{
  
int i,j,k,posa,posb,inia,fina,inib,finb,inik,fink;
double s;



#pragma omp parallel for private(i,j,posa,posb, k, s, inia, fina, inib, finb, inik, fink)
for (i = 0; i < num_rows; i++) 
  {
    for(j = 0; j < t; j++)
    {
      s=0.;
      for(posa=0;filasa[i*(t+2)+posa]!=t;posa+=2)
      {
        for(posb=0;columnasb[j*(t+2)+posb]!=t;posb+=2)
        {
          inia=filasa[i*(t+2)+posa];
          fina=filasa[i*(t+2)+posa+1];
          inib=columnasb[j*(t+2)+posb];
          finb=columnasb[j*(t+2)+posb+1];
          inik=(inia>inib)?inia:inib;
          fink=(fina<finb)?fina:finb;
          if(inia<=finb && fina>=inib)
          {
            for (k = inik; k <=fink; k++)
	            s = s+a[i*t+k]*b[k*t+j];
          }
        }
      }
      c[i*t+j]=s;
    }
  }

}



/**
*
* t: size
* *a: pointer to start of matrix a
* *b: matrix b
* *c: result matrix c
* nodo: id of the node
* np: number of processes
*
**/
void mh(int t,double *a,double *b,double *c,int nodo,int np)
{
  int i,j,k,*filasa,*columnasb,posa,posb,inia,fina,inib,finb,inik,fink;
  double s;
  int num_rows, first_row, fini, ftam;
  
  MPI_Status status;
  
  
  // set thread count
  omp_set_num_threads(8);
  
  
  // broadcast size t to all processes
  MPI_Bcast(&t,1,MPI_INT,0,MPI_COMM_WORLD);
  
  
  // give every process the same amount of rows (block distribution)
  
  // number of rows for the current process
  num_rows = t/np+(t%np>nodo?1:0);
  
  if(nodo!=0)
  {
    // reserve just the space which is needed
    a = (double *) malloc(sizeof(double)*num_rows*t);
    b = (double *) malloc(sizeof(double)*t*t);
    c = (double *) malloc(sizeof(double)*num_rows*t);
  }
  
  
  // broadcast the matrizes to all processes
  // - b for all the same
  // - a row whise in blocks
  if(nodo==0)
  {
    fini= num_rows;
    for(i=1;i<np;i++)
    {
      ftam=t/np+(t%np>i?1:0);
      MPI_Send(&a[t*fini],ftam*t,MPI_DOUBLE,i,20,MPI_COMM_WORLD);
      fini+=ftam;
    }
    MPI_Bcast(b,t*t,MPI_DOUBLE,0,MPI_COMM_WORLD);
  }
  else
  {
    MPI_Recv(a, num_rows*t,MPI_DOUBLE,0,20,MPI_COMM_WORLD,&status);
    MPI_Bcast(b,t*t,MPI_DOUBLE,0,MPI_COMM_WORLD);
  }
  
  
  // reserve space for rows and columns
  filasa= (int *) calloc(sizeof(int), num_rows*(t+2));
  columnasb= (int *) calloc(sizeof(int),t*(t+2));
  
  // get first (starting) row for this process
  if (t%np>nodo)
  {
    first_row = nodo * (t/np + 1);
  }
  else
  {
    first_row = t%np * (t/np + 1) + (nodo - t%np) * t/np;
  }
  
  // preprocessing: get the structure of the matrizes
  // reducing the zero values
  for(i = 0; i < num_rows; i++)
  {
    formarestructura(t,&filasa[i*(t+2)],&a[(i)*t],1);
  }
  
  // this can be calculated by just one process and broadcasted to the others
  // but i think it doesnt make a difference here..
  for(i = 0; i < t; i++)
  {
    formarestructura(t,&columnasb[i*(t+2)],&b[i],t);
  }
  
  
  
#ifdef DEBUG
  printf("STRUCTURE A AND B: \n\n");
  for (i=0; i < num_rows*(t+2); i++) 
  {
    printf("%d ", filasa[i]);
  }
  printf("\n\n");
  
  for (i=0; i < t*(t+2); i++) 
  {
    printf("%d ", columnasb[i]);
  }
  printf("\n\n");
  
  printf("===============================\n\n");
#endif
  
  
  mh_partial(filasa, columnasb, a, b, c, t, num_rows);
  
  // send all results to process master
  if(nodo==0)
  {
    fini=num_rows;
    for(i=1;i<np;i++)
    {
      ftam=t/np+(t%np>i?1:0);
      MPI_Recv(&c[t*fini],ftam*t,MPI_DOUBLE,i,30,MPI_COMM_WORLD,&status);
      fini+=ftam;
    }
  }
  else
  {
    MPI_Send(c,num_rows*t,MPI_DOUBLE,0,30,MPI_COMM_WORLD);
  } 


  free(filasa);
  free(columnasb);
  
  if(nodo!=0)
  {
    free(a);
    free(b);
    free(c);
  }

}

